// Execution thread convention:
// rdi <- binary thread

// Resources used:
// https://mort.coffee/home/fast-interpreters/
// https://blog.reverberate.org/2021/04/21/musttail-efficient-interpreters.html
// https://en.wikibooks.org/wiki/X86_Assembly/GNU_assembly_syntax
// https://www.cs.princeton.edu/courses/archive/spr18/cos217/lectures/15_AssemblyFunctions.pdf
// https://ziglang.org/documentation/master/#toc-Assembly
// https://csiflabs.cs.ucdavis.edu/~ssdavis/50/att-syntax.htm

pub const Word = u64;

// todo: Variant that pushes array of words.
/// (iw | -- iw)
pub fn opPushWord() callconv(.Naked) noreturn {
    asm volatile (
        \\ add $0x10, %%rdi
        \\ pushq -8(%%rdi)
        \\ jmpq *(%%rdi)
    );
}

// todo: Variant that discards array of words.
/// (w --)
pub fn opSinkWord() callconv(.Naked) noreturn {
    asm volatile (
        \\ add $0x08, %%rdi
        \\ addq $0x08, %%rsp
        \\ jmpq *(%%rdi)
    );
}

/// (iw | -- (iw'nth word from stack) )
// fn opTakeWord(binary: [*]const Word, cond: bool) noreturn {
//     @setRuntimeSafety(false);
//     takeWord(binary[1].word);
//     @call(.always_tail, binary[2].function, .{ &binary[2], cond });
// }

/// (iw | w)
// fn opSetWord(binary: [*]const Word, cond: bool) noreturn {
//     @setRuntimeSafety(false);
//     setWord(binary[1].word, popWord());
//     @call(.always_tail, binary[2].function, .{ &binary[2], cond });
// }

// todo: Generate operation permutations procedurally.
// todo: Jump on overflow instead of cond setting?
/// (w1 w2 -- sum overflow)
pub fn opSumWordsWithOverflow() callconv(.Naked) noreturn {
    // https://www.felixcloutier.com/x86/adc
    // https://www.felixcloutier.com/x86/setcc
    // idea: Could https://www.felixcloutier.com/x86/cmovcc be better for overflow push?
    asm volatile (
        \\ movq (%%rsp), %%rax
        \\ adcq 8(%%rsp), %%rax
        \\ movq %%rax, 8(%%rsp)
        \\ setc %%al
        \\ movb %%al, 7(%%rsp)
        \\ addq $0x08, %%rdi
        \\ jmpq *(%%rdi)
    );
}

// todo: Generate operation permutations procedurally.
// todo: We might not need cond register if conditions and jumps are combined?
/// (w1 w2)
// fn opRelativeJumpIfGreaterThan(binary: [*]const Word, cond: bool) noreturn {
//     @setRuntimeSafety(false);
//     const offset = if (popWord() > popWord()) binary[1].word else 2;
//     @call(.always_tail, binary[offset].function, .{ &binary[offset], cond });
// }

/// (addr)
pub fn opReturn() callconv(.Naked) noreturn {
    // https://www.felixcloutier.com/x86/ret
    asm volatile ("ret");
}

pub fn execute(binary: []const Word, entry_addr: usize) void {
    // todo: Ensure correctness.
    // https://wiki.osdev.org/System_V_ABI
    // https://www.felixcloutier.com/x86/call
    asm volatile (
        \\ call *(%%rdi)
        :
        : [thread] "rdi" (&binary[entry_addr]),
        : "rflags", "rax", "rbx", "rsp", "rbp", "r12", "r13", "r14", "r15", "rsi", "rdx", "rcx", "r8", "r9", "r10", "r11", "memory"
    );
}
